/*
    Diablo 2 Character Editor
    Copyright (C) 2000-2001  Burton Tsang

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
//---------------------------------------------------------------------------

#pragma hdrstop

#include "Character.h"
#include "CharacterConstants.h"
//---------------------------------------------------------------------------

#pragma package(smart_init)
//---------------------------------------------------------------------------
Character::Character()
{
   initialize();
}
//---------------------------------------------------------------------------
Character::~Character()
{
   if (isFileOpen()) close();

   if (Items)
      delete[] Items;

   Items = NULL;
   temparray = NULL;
}
//---------------------------------------------------------------------------
/*
   Returns true if there are non-perfect gems.
   Returns false if no gems or all gems are perfect.

   This function is not fully portable as the index values for the items
   are based on little-endian CPUs.  The index values for the outer and
   inner switch expressions may need swapping for big-endian CPUs.
*/
bool Character::anyNonPerfectGems() const
{
   for (int i = 0; i < totalitems; ++i)
   {
      if (MASK_LSB (Version) == 0x40)        // pre-1.07 character file
      {
         if (Items[i].size() == V100_ITEM_LENGTH)  // v1.00 - v1.03
            switch(Items[i][9])
            {
               case 0x13: switch(Items[i][8])
                          {
                             // amethyst
                             case 0xC0:
                             case 0xD0:
                             case 0xE0:
                             case 0xF0: return true;
                          }
                          break;

               case 0x14: switch(Items[i][8])
                          {
                             case 0x10:   // topaz
                             case 0x20:
                             case 0x30:
                             case 0x40:
                             case 0x60:   // sapphire
                             case 0x70:
                             case 0x80:
                             case 0x90:
                             case 0xB0:   // emerald
                             case 0xC0:
                             case 0xD0:
                             case 0xE0: return true;
                          }
                          break;

               case 0x15: switch(Items[i][8])
                          {
                             case 0x00:   // ruby
                             case 0x10:
                             case 0x20:
                             case 0x30:
                             case 0x50:   // diamond
                             case 0x60:
                             case 0x70:
                             case 0x80: return true;
                          }
                          break;

               case 0x16: switch(Items[i][8])
                          {
                             // skull
                             case 0x40:
                             case 0x50:
                             case 0x60:
                             case 0x70: return true;
                          }
                          break;
            }  // end switch
         else              // v1.04 - v1.06
         {
            if (Items[i].size() < 12)
               continue;

            switch(Items[i][11])
            {
               case 0x8D: // chipped
               case 0x99: // flawed
               case 0xCD: // regular
                          switch(Items[i][12])
                          {
                             case 0xD9: // amethyst
                             case 0xE5:
                             case 0x89:
                             case 0x9D:
                             case 0xC9:
                             case 0xDD: // diamond
                                        return true;
                          }
                          break;

               // skull
               case 0xAD: switch(Items[i][12])
                          {
                             case 0x8D: // chipped
                             case 0x99: // flawed
                             case 0xD5: // regular
                             case 0xB1: // flawless
                                        return true;
                          }
                          break;

               // special case, found a flawless amethyst
               case 0xE9: if (Items[i][12] == 0xD9) return true;
                          break;

               case 0xB1: // flawless
                          switch(Items[i][12])
                          {
                             case 0xE5: // topaz
                             case 0x89:
                             case 0x9D:
                             case 0xC9:
                             case 0xDD: // diamond
                                        return true;
                          }
                          break;
            }  // end switch
         }
      }

      // 1.07+ character file including expansion set characters
      else if (MASK_LSB (Version) == 0x50)
      {
         if (Items[i].size() != V107_GEM_LENGTH)
            continue;

         switch(Items[i][10])
         {
            case 0x36: // chipped or regular
            case 0x66: // flawed
            case 0xA6: // flawless amethyst
            case 0xC6: // flawless
                       switch(Items[i][11])
                       {
                          case 0x66:
                          case 0x67: // amethyst

                          case 0x76: // diamond or emerald, depends on 11th value
                          case 0x77: // 7 = diamond; 6 = emerald

                          case 0x26: // ruby or sapphire, depends on 11th value
                          case 0x27: // 7 = ruby; 6 = sapphire

                          case 0x96:
                          case 0x97: // topaz
                                     return true;
                       }
                       break;

            // skull
            case 0xB7: switch(Items[i][11])
                       {
                          case 0x36: // chipped
                          case 0x56: // regular
                          case 0x66: // flawed
                          case 0xC6: // flawless
                                     return true;
                       }
                       break;
         }  // end switch
      }  // end big if
   }  // end for

   return false;
}
//---------------------------------------------------------------------------
bool Character::areSkillsMaxed() const
{
   for (int i = 0; i < NUM_OF_SKILLS; ++i)
      if (Skills[i] != MAX_SKILL_VALUE) return false;

   return true;
}
//---------------------------------------------------------------------------
/*
   Thanks goes to Stoned2000 for making his checksum calculation source
   available to the public.  The Visual Basic source for his editor can be
   found at http://stoned.d2network.com.
*/
void Character::calculateChecksum()
{
   // make sure we start at the beginning of the file
   std::rewind(charfile);

   Checksum = 0;

   unsigned char data, overflow = 0;

   for (unsigned long i = 0; i < FileSize; ++i)
   {
      // doubles the checksum result by left shifting once
      Checksum <<= 1;

      std::fread(&data, sizeof(data), 1, charfile);

      Checksum += data + overflow;

      if (Checksum < 0)
         overflow = 1;
      else
         overflow = 0;
   }

   // rewind the file in case we need to read/write to it again
   std::rewind(charfile);
}
//---------------------------------------------------------------------------
/*
   Sets the correct misc info value
   This function is only called by writeTempFile()
*/
void Character::checkMiscInfo()
{
   if (Experience == 0 && GoldInBelt == 0 && GoldInStash == 0)
      MiscInfo = Nothing;
   else if (Experience > 0 && GoldInBelt == 0 && GoldInStash == 0)
      MiscInfo = ExpOnly;
   else if (Experience == 0 && GoldInBelt > 0 && GoldInStash == 0)
      MiscInfo = BeltOnly;
   else if (Experience > 0 && GoldInBelt > 0 && GoldInStash == 0)
      MiscInfo = ExpBelt;
   else if (Experience == 0 && GoldInBelt == 0 && GoldInStash > 0)
      MiscInfo = StashOnly;
   else if (Experience > 0 && GoldInBelt == 0 && GoldInStash > 0)
      MiscInfo = ExpStash;
   else if (Experience == 0 && GoldInBelt > 0 && GoldInStash > 0)
      MiscInfo = GoldOnly;
   else if (Experience > 0 && GoldInBelt > 0 && GoldInStash > 0)
      MiscInfo = All;
}
//---------------------------------------------------------------------------
void Character::clearLocations()
{
   items_location = quests_location = skills_location = stats_location =
      waypoints_location = 0;
}
//---------------------------------------------------------------------------
void Character::clearSkillChoices()
{
   SkillChoices = 0;
}
//---------------------------------------------------------------------------
void Character::close()
{
   std::fclose(charfile);
   charfile = NULL;
}
//---------------------------------------------------------------------------
/*
   This function should be called only after filling temparray.
*/
void Character::fillItemsArray()
{
   if (Items)
      delete[] Items;

   Items = new Item[totalitems];

   std::fseek(charfile, items_location, SEEK_SET);

   int itemnumber = 0;
   for (int i = 0; i < totalbytes - 1; ++i)
   {
      if (temparray[i] == 0x4A && temparray[i+1] == 0x4D)
      {
         int itemsize = 2;
         int j;
         for (j = i+2; j < totalbytes - 1; ++j)
         {
            if (temparray[j] == 0x4A && temparray[j+1] == 0x4D)
               break;

            ++itemsize;
         }

         if (j == totalbytes - 1)
            ++itemsize;

         Items[itemnumber].size(itemsize);

         for (int k = 0; k < itemsize; ++k)
            Items[itemnumber][k] = temparray[i + k];

         ++itemnumber;
         i += itemsize - 1;
      }  // end if
   }  // end for
}
//---------------------------------------------------------------------------
/*
   Scans the entire character file and marks the location of where quests,
   waypoints, character stats, and skills are located for later use.
   The position of where items are first found in the file is also marked.
*/
void Character::findLocations()
{
   bool foundall = false;
   unsigned char value;
   int cur_pos = std::ftell(charfile);

   std::fseek(charfile, cur_pos, SEEK_SET);
   while(!feof(charfile) && !foundall)
   {
      std::fread(&value, sizeof(value), 1, charfile);
      switch(value)
      {
         case 0x2A: std::fread(&value, sizeof(value), 1, charfile);
                    // found quests marker (0x2A01)
                    if (value == 0x01 && (std::ftell(charfile) > MIN_START_POS))
                       quests_location = std::ftell(charfile);
                    break;

         case 0x57: std::fread(&value, sizeof(value), 1, charfile);
                    // found waypoints marker
                    if (value == 0x53 && (std::ftell(charfile) > MIN_START_POS))
                       waypoints_location = std::ftell(charfile) + 8;

         case 0x67: std::fread(&value, sizeof(value), 1, charfile);
                    // found stats marker (0x6766)
                    if (value == 0x66 && (std::ftell(charfile) > MIN_START_POS))
                       stats_location = std::ftell(charfile);
                    break;

         case 0x69: std::fread(&value, sizeof(value), 1, charfile);
                    // found skills marker (0x6966)
                    if (value == 0x66 && (std::ftell(charfile) > MIN_START_POS))
                    {
                       skills_location = std::ftell(charfile);
                       std::fseek(charfile, NUM_OF_SKILLS+2, SEEK_CUR);

                       std::fread(&NumOfItems, sizeof(NumOfItems), 1, charfile);
                       items_location = std::ftell(charfile);

                       foundall = true;
                    }
                    break;
      }
   }
}
//---------------------------------------------------------------------------
void Character::fillBasicStats(BasicStats& bs)
{
   strcpy(bs.Name, Name);
   bs.Status = getStatus();
   bs.Title = getTitle();
   bs.Class = Class;
   bs.DifficultyLastPlayed = getDifficultyLastPlayed();
   bs.StartingAct = getStartingAct();
}
//---------------------------------------------------------------------------
void Character::fillCharacterStats(CharacterStats& cs)
{
   cs.Strength = Strength;
   cs.Energy = Energy;
   cs.Dexterity = Dexterity;
   cs.Vitality = Vitality;

   cs.CurLife = CurLife;
   cs.MaxLife = MaxLife;
   cs.CurMana = CurMana;
   cs.MaxMana = MaxMana;
   cs.CurStamina = CurStamina;
   cs.MaxStamina = MaxStamina;

   cs.Level = Level;
   cs.Experience = Experience;
   cs.GoldInBelt = GoldInBelt;
   cs.GoldInStash = GoldInStash;
}
//---------------------------------------------------------------------------
char Character::getClass() const
{
   return Class;
}
//---------------------------------------------------------------------------
/*
   Returns a value indicating the difficulty level the character last played at,
   where 0 = normal, 1 = nightmare, 2 = hell
*/
char Character::getDifficultyLastPlayed()
{
   if (Version < VERSION_109)
      return DifficultyAndAct >> 4;
   else
   {
      if (StartingAct[0] != 0)
         return NORMAL;
      else if (StartingAct[1] != 0)
         return NIGHTMARE;
      else
         return HELL;
   }
}
//---------------------------------------------------------------------------
unsigned long Character::getExperience() const
{
   return Experience;
}
//---------------------------------------------------------------------------
long Character::getLevel() const
{
   return Level;
}
//---------------------------------------------------------------------------
char *Character::getName(bool fromfile)
{
   if (fromfile)
   {
      if (Version < VERSION_109)
         std::fseek(charfile, 8, SEEK_SET);
      else
         std::fseek(charfile, 20, SEEK_SET);

      std::fread(Name, sizeof(Name), 1, charfile);
   }

   return Name;
}
//---------------------------------------------------------------------------
QuestInfo *Character::getQuests()
{
   return Difficulty;
}
//---------------------------------------------------------------------------
long Character::getSkillChoices() const
{
   return SkillChoices;
}
//---------------------------------------------------------------------------
char *Character::getSkills()
{
   return Skills;
}
//---------------------------------------------------------------------------
/*
   Returns a value indicating the act the character was saved at, where 0
   indicates the first act
*/
char Character::getStartingAct()
{
   if (Version < VERSION_109)
      return MASK_MSB (DifficultyAndAct);
   else
   {
      if (StartingAct[0] != 0)
         return MASK_MSB (StartingAct[0]);
      else if (StartingAct[1] != 0)
         return MASK_MSB (StartingAct[1]);
      else
         return MASK_MSB (StartingAct[2]);
   }
}
//---------------------------------------------------------------------------
char Character::getStatus(bool rawvalue) const
{
   if (rawvalue)
      return Status;

   if (MASK_MSB (Status) < Hardcore)
      return NoDeaths;
   else if (MASK_MSB (Status) >= Hardcore && MASK_MSB (Status) < Resurrected)
      return Hardcore;
   else if (MASK_MSB (Status) >= Resurrected && MASK_MSB (Status) < Dead)
      return Resurrected;
   else
      return Dead;
}
//---------------------------------------------------------------------------
char Character::getTitle(bool rawvalue) const
{
   if (rawvalue)
      return Title;

   switch(Title)
   {
      case SirDame:
      case Slayer:
      case 0x06:
      case 0x07: if (!isExpansionCharacter())
                    return SirDame;
                 else
                    return Slayer;
      case LordLady:
      case Champion:
      case 0x09:
      case 0x0B: if (!isExpansionCharacter())
                    return LordLady;
                 else
                    return Champion;
      case BaronBaroness:
      case MPatriarch:
      case 0x0D:
      case 0x0E: if (!isExpansionCharacter())
                    return BaronBaroness;
                 else
                    return MPatriarch;
   }

   return None;
}
//---------------------------------------------------------------------------
/*
   Returns the total number of items in temparray.
   This function should only be called after filling temparray.
*/
unsigned short Character::getTotalNumberOfItems() const
{
   int totalitems = 0;

   for (int i = 0; i < totalbytes - 1; ++i)
      if (temparray[i] == 0x4A && temparray[i+1] == 0x4D)
      {
         ++totalitems;
         ++i;
      }

   return totalitems;
}
//---------------------------------------------------------------------------
long Character::getVersion() const
{
   return Version;
}
//---------------------------------------------------------------------------
unsigned long Character::getWaypoints(int difficulty, int setnumber) const
{
   return Waypoints[difficulty][setnumber];
}
//---------------------------------------------------------------------------
void Character::initialize()
{
   std::memset(Header, 0, sizeof(Header));
   std::memset(Name, 0, sizeof(Name));
   Status = 0;
   Title = 0;
   Class = 0;

   DifficultyAndAct = 0;
   std::memset(StartingAct, 0, sizeof(StartingAct));
   std::memset(Difficulty, 0, sizeof(Difficulty));
   std::memset(Waypoints, 0, sizeof(Waypoints));

   StatsSkillsInfo = 0;
   MiscInfo = 0;
   Dexterity = 0;
   Energy = 0;
   Level = DisplayLevel = 1;
   Strength = 0;
   Vitality = 0;

   CurLife = 0;
   MaxLife = 0;
   CurMana = 0;
   MaxMana = 0;
   CurStamina = 0;
   MaxStamina = 0;

   Experience = 0;
   GoldInBelt = 0;
   GoldInStash = 0;

   std::memset(Skills, 0, sizeof(Skills));
   Items = NULL;

   temparray = NULL;
   totalbytes = 0;

   clearLocations();
}
//---------------------------------------------------------------------------
bool Character::isExpansionCharacter() const
{
   if (MASK_LSB (Status) == EXPANSION_CHARACTER)
      return true;
   else
      return false;
}
//---------------------------------------------------------------------------
bool Character::isFileOpen() const
{
   if (charfile)
      return true;
   else
      return false;
}
//---------------------------------------------------------------------------
/*
   Returns true if the file has a valid header
*/
bool Character::isValidHeader() const
{
   int result = std::memcmp(Header, HEADER, HEADER_LENGTH);

   if (result == 0)
      return true;
   else
      return false;
}
//---------------------------------------------------------------------------
void Character::maxSkills()
{
   std::memset(Skills, MAX_SKILL_VALUE, sizeof(Skills));
}
//---------------------------------------------------------------------------
/*
   Returns false if file was not opened or there was an error.
*/
bool Character::open(const char *filename)
{
   charfile = std::fopen(filename, "rb+");
   this->filename = filename;

   if (charfile == NULL)
      return false;

   readHeader();
   return true;
}
//---------------------------------------------------------------------------
void Character::read(bool update_locations)
{
   if (update_locations)
   {
      initialize();
      findLocations();
   }

   readBasicInfo();
   readQuests();
   readWaypoints();
   readStats();
   readSkills();
   readItems();
}
//---------------------------------------------------------------------------
void Character::readBasicInfo()
{
   std::fseek(charfile, 4, SEEK_SET);
   std::fread(&Version, sizeof(Version), 1, charfile);

   if (Version >= VERSION_109)
   {
      std::fread(&FileSize, sizeof(FileSize), 1, charfile);
      std::fread(&Checksum, sizeof(Checksum), 1, charfile);

      std::fseek(charfile, 20, SEEK_SET);
   }

   std::fread(Name, sizeof(Name), 1, charfile);
   std::fread(&Status, sizeof(Status), 1, charfile);
   std::fread(&Title, sizeof(Title), 1, charfile);

   if (Version < VERSION_109)
      std::fseek(charfile, 34, SEEK_SET);
   else
      std::fseek(charfile, 40, SEEK_SET);

   std::fread(&Class, sizeof(Class), 1, charfile);

   if (Version < VERSION_109)
      std::fseek(charfile, 36, SEEK_SET);
   else
      std::fseek(charfile, 43, SEEK_SET);

   std::fread(&DisplayLevel, sizeof(DisplayLevel), 1, charfile);

   if (Version < VERSION_109)
   {
      std::fseek(charfile, 88, SEEK_SET);
      std::fread(&DifficultyAndAct, sizeof(DifficultyAndAct), 1, charfile);
   }
   else
   {
      std::fseek(charfile, 168, SEEK_SET);
      std::fread(StartingAct, sizeof(StartingAct), 1, charfile);
   }

}
//---------------------------------------------------------------------------
void Character::readHeader()
{
   std::rewind(charfile);
   std::fread(Header, sizeof(Header), 1, charfile);
}
//---------------------------------------------------------------------------
void Character::readItems()
{
   std::fseek(charfile, 0, SEEK_END);
   totalbytes = std::ftell(charfile) - items_location;

   temparray = new unsigned char[totalbytes];

   std::fseek(charfile, items_location, SEEK_SET);
   std::fread(temparray, totalbytes, 1, charfile);

   totalitems = getTotalNumberOfItems(); 
   fillItemsArray();

   delete[] temparray;
}
//---------------------------------------------------------------------------
void Character::readQuests()
{
   std::fseek(charfile, quests_location, SEEK_SET);
   for (int i = 0; i < NUM_OF_DIFFICULTY; ++i)
   {
      for (int j = 0; j < NUM_OF_ACTS - 1; ++j)
      {
         std::fread(&Difficulty[i].ActIntro[j], sizeof(Difficulty[i].ActIntro[j]),
                    1, charfile);

         for (int k = 0; k < NUM_OF_QUESTS; ++k)
            std::fread(&Difficulty[i].Quests[j][k],
                       sizeof(Difficulty[i].Quests[j][k]), 1, charfile);

         if (j < 3)
            std::fread(&Difficulty[i].ActCompleted[j],
                       sizeof(Difficulty[i].ActCompleted[j]), 1, charfile);
      }

      std::fseek(charfile, -6, SEEK_CUR);
      std::fread(&Difficulty[i].ActCompleted[3], sizeof(Difficulty[i].ActCompleted[3]),
                 1, charfile);

      // move on if pre-1.07 version
      if (MASK_LSB (Version) == 0x40)
         std::fseek(charfile, 28, SEEK_CUR);

      // read act 5 quests
      else
      {
         std::fseek(charfile, 12, SEEK_CUR);

         for (int k = 0; k < NUM_OF_QUESTS; ++k)
            std::fread(&Difficulty[i].Quests[4][k],
                       sizeof(Difficulty[i].Quests[4][k]), 1, charfile);

         std::fseek(charfile, 14, SEEK_CUR);
      }
   }
}
//---------------------------------------------------------------------------
void Character::readSkills()
{
   std::fseek(charfile, skills_location, SEEK_SET);
   std::fread(Skills, sizeof(Skills), 1, charfile);
}
//---------------------------------------------------------------------------
void Character::readStats()
{
   // find out if there is any new stats and skill choices remaining
   std::fseek(charfile, stats_location, SEEK_SET);
   std::fread(&StatsSkillsInfo, sizeof(StatsSkillsInfo), 1, charfile);

   // is there any experience and/or gold
   std::fread(&MiscInfo, sizeof(MiscInfo), 1, charfile);


   // skip null byte character version is less than 1.09
   if (Version < VERSION_109)
      std::fseek(charfile, 1, SEEK_CUR);

   // read basic stats
   std::fread(&Strength, sizeof(Strength), 1, charfile);
   std::fread(&Energy, sizeof(Energy), 1, charfile);
   std::fread(&Dexterity, sizeof(Dexterity), 1, charfile);
   std::fread(&Vitality, sizeof(Vitality), 1, charfile);

   switch(StatsSkillsInfo)
   {
      case StatsOnly: std::fread(&StatsLeft, sizeof(StatsLeft), 1, charfile);
                      break;
      case SkillsOnly: std::fread(&SkillChoices, sizeof(SkillChoices), 1, charfile);
                       break;
      case StatsSkills: std::fread(&StatsLeft, sizeof(StatsLeft), 1, charfile);
                        std::fread(&SkillChoices, sizeof(SkillChoices), 1, charfile);
                        break;
   }

   // skip null byte
   std::fseek(charfile, 1, SEEK_CUR);

   if (StatsSkillsInfo != NoCurLife)
   {
      std::fread(&CurLife, sizeof(CurLife), 1, charfile);
      std::fseek(charfile, 2, SEEK_CUR);
   }

   std::fread(&MaxLife, sizeof(MaxLife), 1, charfile);

   std::fseek(charfile, 2, SEEK_CUR);
   std::fread(&CurMana, sizeof(CurMana), 1, charfile);
   std::fseek(charfile, 2, SEEK_CUR);
   std::fread(&MaxMana, sizeof(MaxMana), 1, charfile);

   std::fseek(charfile, 2, SEEK_CUR);
   std::fread(&CurStamina, sizeof(CurStamina), 1, charfile);
   std::fseek(charfile, 2, SEEK_CUR);
   std::fread(&MaxStamina, sizeof(MaxStamina), 1, charfile);

   std::fseek(charfile, 1, SEEK_CUR);
   std::fread(&Level, sizeof(Level), 1, charfile);

   switch(MiscInfo)
   {
      case ExpOnly: std::fread(&Experience, sizeof(Experience), 1, charfile);
                    break;

      case BeltOnly: std::fread(&GoldInBelt, sizeof(GoldInBelt), 1, charfile);
                     break;

      case ExpBelt: std::fread(&Experience, sizeof(Experience), 1, charfile);
                    std::fread(&GoldInBelt, sizeof(GoldInBelt), 1, charfile);
                    break;

      case StashOnly: std::fread(&GoldInStash, sizeof(GoldInStash), 1, charfile);
                      break;

      case ExpStash: std::fread(&Experience, sizeof(Experience), 1, charfile);
                     std::fread(&GoldInStash, sizeof(GoldInStash), 1, charfile);
                     break;

      case All: std::fread(&Experience, sizeof(Experience), 1, charfile);
      case GoldOnly: std::fread(&GoldInBelt, sizeof(GoldInBelt), 1, charfile);
                std::fread(&GoldInStash, sizeof(GoldInStash), 1, charfile);
                break;
   }
}
//---------------------------------------------------------------------------
void Character::readWaypoints()
{
   std::fseek(charfile, waypoints_location, SEEK_SET);

   for (int i = 0; i < NUM_OF_DIFFICULTY; ++i)
   {
      std::fread(Waypoints[i], sizeof(Waypoints[i]), 1, charfile);
      std::fseek(charfile, 16, SEEK_CUR);
   }
}
//---------------------------------------------------------------------------
void Character::save()
{
   writeBasicStats();
   writeQuests();
   writeWaypoints();
   writeStats();
   writeSkills();
   writeItems();

   writeTempFile();

   // prepare to update the character file
   std::fclose(charfile);
   std::remove(filename.c_str());

   // check to see if the filename needs to be changed
   // to match the character's name
   unsigned int pos = filename.rfind('\\', filename.length());

   if (pos != filename.length())
      ++pos;

   std::string tempname = filename;
   tempname.erase(0, pos);

   // compare filename (w/o extension) to character's name
   if (tempname.compare(0, tempname.length()-4, Name) != 0)
   {
      filename.erase(pos, tempname.length());
      tempname = Name;
      tempname += ".d2s";
      filename.insert(filename.length(), tempname);
   }

   // rename temp file to character file
   std::rename(tempfilename.c_str(), filename.c_str());
   open(filename.c_str());

   // determine if a checksum needs to be calculated and stored
   if (Version >= VERSION_109)
   {
      // store the file's size
      std::fseek(charfile, 0, SEEK_END);
      FileSize = std::ftell(charfile);
      std::fseek(charfile, 8, SEEK_SET);
      std::fwrite(&FileSize, sizeof(FileSize), 1, charfile);

      // make sure the checksum is zero in the file
      Checksum = 0;
      std::fwrite(&Checksum, sizeof(Checksum), 1, charfile);

      calculateChecksum();

      // write the checksum into the file
      std::fseek(charfile, 12, SEEK_SET);
      std::fwrite(&Checksum, sizeof(Checksum), 1, charfile);
   }
}
//---------------------------------------------------------------------------
void Character::setWaypoints(int difficulty, int setnumber, unsigned long newvalue)
{
   Waypoints[difficulty][setnumber] = newvalue;
}
//---------------------------------------------------------------------------
void Character::updateBasicStats(const BasicStats& bs)
{
   strcpy(Name, bs.Name);
   Status = MASK_LSB (Status) | bs.Status * Hardcore;

   if (!isExpansionCharacter())
   {
      if (bs.Title * SirDame != getTitle())
         Title = bs.Title * SirDame;
   }
   else
   {
      if (bs.Title * Slayer != getTitle())
         Title = bs.Title * Slayer;
   }

   Class = bs.Class;

   if (Version < VERSION_109)
      DifficultyAndAct = 0x10 * bs.DifficultyLastPlayed | bs.StartingAct;
   else
   {
      std::memset(StartingAct, 0, sizeof(StartingAct));
      StartingAct[bs.DifficultyLastPlayed] = 0x80 | bs.StartingAct;
   }
}
//---------------------------------------------------------------------------
void Character::updateCharacterStats(const CharacterStats& cs)
{
   Dexterity = cs.Dexterity;
   Energy = cs.Energy;
   Strength = cs.Strength;
   Vitality = cs.Vitality;

   CurLife = cs.CurLife;
   MaxLife = cs.MaxLife;
   CurMana = cs.CurMana;
   MaxMana = cs.MaxMana;
   CurStamina = cs.CurStamina;
   MaxStamina = cs.MaxStamina;

   Level = cs.Level;
   DisplayLevel = Level;         // updates character's display level
                                 // to match character's actual level
   Experience = cs.Experience;
   GoldInBelt = cs.GoldInBelt;
   GoldInStash = cs.GoldInStash;
}
//---------------------------------------------------------------------------
void Character::updateQuests(const QuestInfo *qi)
{
   std::memcpy(Difficulty, qi, sizeof(Difficulty));
}
//---------------------------------------------------------------------------
void Character::updateSkills(const char *updated_skills)
{
   std::memcpy(Skills, updated_skills, sizeof(updated_skills));
}
//---------------------------------------------------------------------------
/*
   This function is not fully portable as the index values for the items
   are based on little-endian CPUs.  The index values for the outer and
   inner switch expressions may need swapping for big-endian CPUs.
*/
void Character::upgradeGems()
{
   for (int i = 0; i < totalitems; ++i)
   {
      if (MASK_LSB (Version) == 0x40)     // pre-1.07 character file
      {
         if (Items[i].size() == V100_ITEM_LENGTH)  // v1.00 - v1.03
            switch(Items[i][9])
            {
               case 0x13: switch(Items[i][8])
                          {
                             // amethyst
                             case 0xC0:
                             case 0xD0:
                             case 0xE0:
                             case 0xF0: Items[i][8] = 0x00;
                                        Items[i][9] = 0x14;
                                        break;
                          }
                          break;

               case 0x14: switch(Items[i][8])
                          {
                             // topaz
                             case 0x10:
                             case 0x20:
                             case 0x30:
                             case 0x40: Items[i][8] = 0x50;
                                        break;
                             // sapphire
                             case 0x60:
                             case 0x70:
                             case 0x80:
                             case 0x90: Items[i][8] = 0xA0;
                                        break;
                             // emerald
                             case 0xB0:
                             case 0xC0:
                             case 0xD0:
                             case 0xE0: Items[i][8] = 0xF0;
                                        break;
                          }
                          break;

               case 0x15: switch(Items[i][8])
                          {
                             // ruby
                             case 0x00:
                             case 0x10:
                             case 0x20:
                             case 0x30: Items[i][8] = 0x40;
                                        break;
                             // diamond
                             case 0x50:
                             case 0x60:
                             case 0x70:
                             case 0x80: Items[i][8] = 0x90;
                                        break;
                          }
                          break;

               case 0x16: switch(Items[i][8])
                          {
                             // skull
                             case 0x40:
                             case 0x50:
                             case 0x60:
                             case 0x70: Items[i][8] = 0x80;
                                        break;
                          }
                          break;
            }  // end switch
         else        // v1.04 - v1.06
         {
            if (Items[i].size() < 12)
               continue;

            switch(Items[i][11])
            {
               case 0x8D: // chipped
               case 0x99: // flawed
               case 0xCD: // regular
                          switch(Items[i][12])
                          {
                             case 0xD9: // amethyst
                             case 0xE5:
                             case 0x89:
                             case 0x9D:
                             case 0xC9:
                             case 0xDD: // diamond
                                        Items[i][11] = 0xC1;
                          }
                          break;

               // skull
               case 0xAD: switch(Items[i][12])
                          {
                             case 0x8D: // chipped
                             case 0x99: // flawed
                             case 0xD5: // regular
                             case 0xB1: // flawless
                                        Items[i][12] = 0xE9;
                          }
                          break;

               // special case, found a flawless amethyst
               case 0xE9: if (Items[i][12] == 0xD9)
                             Items[i][11] = 0xC1;
                          break;

               case 0xB1: // flawless
                          switch(Items[i][12])
                          {
                             case 0xE5: // topaz
                             case 0x89:
                             case 0x9D:
                             case 0xC9:
                             case 0xDD: // diamond
                                        Items[i][11] = 0xC1;
                          }
                          break;
            }  // end switch
         }  // end else
      }

      // 1.07+ character file including expansion set characters
      else if (MASK_LSB (Version) == 0x50)
      {
         if (Items[i].size() != V107_GEM_LENGTH)
            continue;

         switch(Items[i][10])
         {
            case 0x36: // chipped or regular
            case 0x66: // flawed
            case 0xA6: // flawless amethyst
            case 0xC6: // flawless
                       switch(Items[i][11])
                       {
                          case 0x66:
                          case 0x67: // amethyst

                          case 0x76: // diamond or emerald, depends on 11th value
                          case 0x77: // 7 = diamond; 6 = emerald

                          case 0x26: // ruby or sapphire, depends on 11th value
                          case 0x27: // 7 = ruby; 6 = sapphire

                          case 0x96:
                          case 0x97: // topaz
                                     Items[i][10] = 0x06;
                                     Items[i][11] |= 0x01;
                                     break;
                       }
                       break;

            // skull
            case 0xB7: switch(Items[i][11])
                       {
                          case 0x36: // chipped
                          case 0x56: // regular
                          case 0x66: // flawed
                          case 0xC6: // flawless
                                     Items[i][11] = 0xA6;
                                     Items[i][12] = 0x07;
                                     break;
                       }
                       break;
         }  // end switch
      }
   }  // end for
}
//---------------------------------------------------------------------------
void Character::writeBasicStats()
{
   writeName();
   std::fwrite(&Status, sizeof(Status), 1, charfile);
   std::fwrite(&Title, sizeof(Title), 1, charfile);

   if (Version < VERSION_109)
      std::fseek(charfile, 34, SEEK_SET);
   else
      std::fseek(charfile, 40, SEEK_SET);

   std::fwrite(&Class, sizeof(Class), 1, charfile);

   if (Version < VERSION_109)
      std::fseek(charfile, 36, SEEK_SET);
   else
      std::fseek(charfile, 43, SEEK_SET);

   std::fwrite(&DisplayLevel, sizeof(DisplayLevel), 1, charfile);

   if (Version < VERSION_109)
   {
      std::fseek(charfile, 88, SEEK_SET);
      std::fwrite(&DifficultyAndAct, sizeof(DifficultyAndAct), 1, charfile);
   }
   else
   {
      std::fseek(charfile, 168, SEEK_SET);
      std::fwrite(StartingAct, sizeof(StartingAct), 1, charfile);
   }
}
//---------------------------------------------------------------------------
void Character::writeItems()
{
   std::fseek(charfile, items_location, SEEK_SET);
   for (int i = 0; i < totalitems; ++i)
      std::fwrite(Items[i].getItemArray(), Items[i].size(), 1, charfile);
}
//---------------------------------------------------------------------------
void Character::writeName()
{
   // make sure name is saved correctly
   std::string temp(Name);
   std::memset(Name, 0, sizeof(Name));
   strcpy(Name, temp.c_str());

   if (Version < VERSION_109)
      std::fseek(charfile, 8, SEEK_SET);
   else
      std::fseek(charfile, 20, SEEK_SET);

   std::fwrite(Name, sizeof(Name), 1, charfile);
}
//---------------------------------------------------------------------------
void Character::writeQuests()
{
   std::fseek(charfile, quests_location, SEEK_SET);
   for (int i = 0; i < NUM_OF_DIFFICULTY; ++i)
   {
      // write values of acts 1 through 4
      for (int j = 0; j < NUM_OF_ACTS - 1; ++j)
      {
         std::fwrite(&Difficulty[i].ActIntro[j], sizeof(Difficulty[i].ActIntro[j]),
                     1, charfile);
         std::fwrite(&Difficulty[i].Quests[j], sizeof(Difficulty[i].Quests[j]),
                     1, charfile);

         if (j < 3)
            std::fwrite(&Difficulty[i].ActCompleted[j],
                        sizeof(Difficulty[i].ActCompleted[j]), 1, charfile);
      }

      // fill in act completed values for act 4
      std::fseek(charfile, -6, SEEK_CUR);
      std::fwrite(&Difficulty[i].ActCompleted[3],
                  sizeof(Difficulty[i].ActCompleted[3]), 1, charfile);


      // move on if pre-1.07 version
      if (MASK_LSB (Version) == 0x40)
         std::fseek(charfile, 28, SEEK_CUR);

      // write act 5 quests info
      else
      {
         std::fseek(charfile, 12, SEEK_CUR);
         std::fwrite(&Difficulty[i].Quests[4], sizeof(Difficulty[i].Quests[4]),
                     1, charfile);
         std::fseek(charfile, 14, SEEK_CUR);
      }
   }
}
//---------------------------------------------------------------------------
void Character::writeSkills()
{
   std::fseek(charfile, skills_location, SEEK_SET);
   std::fwrite(Skills, sizeof(Skills), 1, charfile);
}
//---------------------------------------------------------------------------
void Character::writeStats()
{
   std::fseek(charfile, stats_location+1, SEEK_SET);
   std::fwrite(&MiscInfo, sizeof(MiscInfo), 1, charfile);

   // skip null byte if character version is less than 1.09
   if (Version < VERSION_109)
      std::fseek(charfile, 1, SEEK_CUR);

   std::fwrite(&Strength, sizeof(Strength), 1, charfile);
   std::fwrite(&Energy, sizeof(Energy), 1, charfile);
   std::fwrite(&Dexterity, sizeof(Dexterity), 1, charfile);
   std::fwrite(&Vitality, sizeof(Vitality), 1, charfile);

   switch(StatsSkillsInfo)
   {
      case StatsOnly: std::fseek(charfile, 4, SEEK_CUR);
                      break;

      case StatsSkills: std::fseek(charfile, 4, SEEK_CUR);
      case SkillsOnly: std::fwrite(&SkillChoices, sizeof(SkillChoices), 1, charfile);
                       break;
   }

   // skip null byte
   std::fseek(charfile, 1, SEEK_CUR);

   if (StatsSkillsInfo != NoCurLife)
   {
      std::fwrite(&CurLife, sizeof(CurLife), 1, charfile);
      std::fseek(charfile, 2, SEEK_CUR);
   }

   std::fwrite(&MaxLife, sizeof(MaxLife), 1, charfile);

   std::fseek(charfile, 2, SEEK_CUR);
   std::fwrite(&CurMana, sizeof(CurMana), 1, charfile);

   std::fseek(charfile, 2, SEEK_CUR);
   std::fwrite(&MaxMana, sizeof(MaxMana), 1, charfile);

   std::fseek(charfile, 2, SEEK_CUR);
   std::fwrite(&CurStamina, sizeof(CurStamina), 1, charfile);

   std::fseek(charfile, 2, SEEK_CUR);
   std::fwrite(&MaxStamina, sizeof(MaxStamina), 1, charfile);

   // skip null byte
   std::fseek(charfile, 1, SEEK_CUR);

   std::fwrite(&Level, sizeof(Level), 1, charfile);

   switch(MiscInfo)
   {
      case ExpOnly: std::fwrite(&Experience, sizeof(Experience), 1, charfile);
                    break;

      case BeltOnly: std::fwrite(&GoldInBelt, sizeof(GoldInBelt), 1, charfile);
                     break;

      case ExpBelt: std::fwrite(&Experience, sizeof(Experience), 1, charfile);
                    std::fwrite(&GoldInBelt, sizeof(GoldInBelt), 1, charfile);
                    break;

      case StashOnly: std::fwrite(&GoldInStash, sizeof(GoldInStash), 1, charfile);
                      break;

      case ExpStash: std::fwrite(&Experience, sizeof(Experience), 1, charfile);
                     std::fwrite(&GoldInStash, sizeof(GoldInStash), 1, charfile);
                     break;

      case GoldOnly: std::fwrite(&GoldInBelt, sizeof(GoldInBelt), 1, charfile);
                     std::fwrite(&GoldInStash, sizeof(GoldInStash), 1, charfile);
                     break;

      case All: std::fwrite(&Experience, sizeof(Experience), 1, charfile);
                std::fwrite(&GoldInBelt, sizeof(GoldInBelt), 1, charfile);
                std::fwrite(&GoldInStash, sizeof(GoldInStash), 1, charfile);
                break;
   }

   std::fwrite(SKILLS_MARKER, sizeof(SKILLS_MARKER), 1, charfile);
   std::fwrite(Skills, sizeof(Skills), 1, charfile);
}
//---------------------------------------------------------------------------
/*
   This function makes sure that any changes to the character's
   experience and gold values are correctly stored in the file.
*/
void Character::writeTempFile()
{
   tempfilename = std::tmpnam(NULL);
   std::FILE *tempfile = std::fopen(tempfilename.c_str(), "wb");

   std::rewind(charfile);

   int startpos = stats_location+1;
   unsigned char filedata;

   while(std::ftell(charfile) < startpos)
   {
      std::fread(&filedata, sizeof(filedata), 1, charfile);
      std::fwrite(&filedata, sizeof(filedata), 1, tempfile);
   }

   unsigned char prevMiscInfo = MiscInfo;

   checkMiscInfo();

   std::fwrite(&MiscInfo, sizeof(MiscInfo), 1, tempfile);
   std::fseek(charfile, 1, SEEK_CUR);

   int numOfBytes = 16+ 1+ 22+1+ 4;

   if (Version < VERSION_109)
      numOfBytes += 1;

   // calculate exact # of bytes to read
   switch(StatsSkillsInfo)
   {
      case StatsOnly:
      case SkillsOnly: numOfBytes += 4;
                       break;
      case StatsSkills: numOfBytes += 4+4;
                        break;
   }

   // read and write # of bytes calculated above
   for (int i = 0; i < numOfBytes; ++i)
   {
      std::fread(&filedata, sizeof(filedata), 1, charfile);
      std::fwrite(&filedata, sizeof(filedata), 1, tempfile);
   }

   // write any experience and/or gold
   switch(MiscInfo)
   {
      case ExpOnly: std::fwrite(&Experience, sizeof(Experience), 1, tempfile);
                    break;

      case BeltOnly: std::fwrite(&GoldInBelt, sizeof(GoldInBelt), 1, tempfile);
                     break;

      case ExpBelt: std::fwrite(&Experience, sizeof(Experience), 1, tempfile);
                    std::fwrite(&GoldInBelt, sizeof(GoldInBelt), 1, tempfile);
                    break;

      case StashOnly: std::fwrite(&GoldInStash, sizeof(GoldInStash), 1, tempfile);
                      break;

      case ExpStash: std::fwrite(&Experience, sizeof(Experience), 1, tempfile);
                     std::fwrite(&GoldInStash, sizeof(GoldInStash), 1, tempfile);
                     break;

      case GoldOnly: std::fwrite(&GoldInBelt, sizeof(GoldInBelt), 1, tempfile);
                     std::fwrite(&GoldInStash, sizeof(GoldInStash), 1, tempfile);
                     break;

      case All: std::fwrite(&Experience, sizeof(Experience), 1, tempfile);
                std::fwrite(&GoldInBelt, sizeof(GoldInBelt), 1, tempfile);
                std::fwrite(&GoldInStash, sizeof(GoldInStash), 1, tempfile);
                break;
   }

   // skip over correct # of bytes in original file
   switch(prevMiscInfo)
   {
      case ExpOnly:
      case BeltOnly:
      case StashOnly: std::fseek(charfile, 4, SEEK_CUR);
                      break;

      case ExpBelt:
      case ExpStash:
      case GoldOnly: std::fseek(charfile, 8, SEEK_CUR);
                     break;

      case All: std::fseek(charfile, 12, SEEK_CUR);
                break;
   }

   std::fwrite(SKILLS_MARKER, sizeof(SKILLS_MARKER), 1, tempfile);
   std::fwrite(Skills, sizeof(Skills), 1, tempfile);
   std::fwrite(ITEM_MARKER, sizeof(ITEM_MARKER), 1, tempfile);
   std::fwrite(&NumOfItems, sizeof(NumOfItems), 1, tempfile);

   // write items
   for (int i = 0; i < totalitems; ++i)
      std::fwrite(Items[i].getItemArray(), Items[i].size(), 1, tempfile);

   std::fclose(tempfile);
}
//---------------------------------------------------------------------------
void Character::writeWaypoints()
{
   std::fseek(charfile, waypoints_location, SEEK_SET);

   for (int i = 0; i < NUM_OF_DIFFICULTY; ++i)
   {
      std::fwrite(Waypoints[i], sizeof(Waypoints[i]), 1, charfile);
      std::fseek(charfile, 16, SEEK_CUR);
   }
}
//---------------------------------------------------------------------------

